@@ -1,10 +1,9 @@
 package FSA::Rules;
 
-# $Id: Rules.pm 2711 2006-03-03 00:05:37Z theory $
+# $Id: Rules.pm 3025 2006-07-14 18:49:52Z theory $
 
 use strict;
-use Clone qw/clone/;
-$FSA::Rules::VERSION = '0.24';
+$FSA::Rules::VERSION = '0.25';
 
 =begin comment
 
@@ -160,6 +159,16 @@ The supported keys in the state definition hash references are:
 
 =over
 
+=item label
+
+  label => 'Do we have a username?',
+  label => 'Create a new user',
+
+A label for the state. It might be the question that is being asked within the
+state (think decsision tree), the answer to which determins which rule will
+trigger the switch to the next state. Or it might merely describe what's
+happening in the state.
+
 =item on_enter
 
   on_enter => sub { ... }
@@ -304,7 +313,6 @@ sub new {
         stack  => [],
         table  => {},
         self   => $self,
-        graph  => clone(\@_),
     };
 
     $params->{state_class} ||= 'FSA::State';
@@ -326,7 +334,7 @@ sub new {
         # Create the state object and cache the state data.
         my $obj = $params->{state_class}->new;
         $def->{name} = $state;
-        $def->{machine} = "$self";
+        $def->{machine} = $self;
         $fsa->{table}{$state} = $obj;
         push @{$fsa->{ord}}, $obj;
         $states{$obj} = $def;
@@ -962,80 +970,127 @@ sub stacktrace {
 =head3 graph
 
   my $graph_viz = $fsa->graph(@graph_viz_args);
-  $graph_viz = $fas->(\%params, @graph_viz_args);
+  $graph_viz = $fsa->graph(\%params, @graph_viz_args);
 
 Constructs and returns a L<GraphViz|GraphViz> object useful for generating
 graphical representations of the complete rules engine. The parameters to
 C<graph()> are all those supported by the GraphViz constructor; consult the
 L<GraphViz|GraphViz> documentation for details.
 
-An optional hash ref of parameters may be passed as the first argument to
-C<graph()>. The supported parameters are:
+Each node in the graph represents a single state. The label for each node in
+the graph will be either the state label or if there is no lable, the state
+name.
+
+Each edge in the graph represents a rule that defines the relationship between
+two states. If a rule is specified as a hash reference, the C<message> key
+will be used as the edge label; otherwise the label will be blank.
+
+An optional hash reference of parameters may be passed as the first argument
+to C<graph()>. The supported parameters are:
 
 =over
 
-=item text_wrap
+=item with_state_name
 
-The text wrap length for graphs.  There's a lot of text on the graph.  This
-property will set the wrap length for all text to the given length using the
-C<Text::Wrap> module.
-
-Each edge on the graph has a "label." If the rules for a given state were
-specified as hash references in the call to C<new()>, the C<message> key will
-be used as the label; otherwise the label is blank. When used as labels,
-messages are wrapped in order to make labels fit better. The default maximum
-line length is 25. However, you may set a different line length using this
-parameter.
-
-B<Note:> By default, text wrapping for graphs is disabled. You must explicitly
-specify what text you want wrapped with either the C<wrap_nodes> or
-C<wrap_labels> parameters.
+This parameter, if set to true, prepends the name of the state and two
+newlines to the label for each node. If a state has no label, then the state
+name is simply used, regardless. Defaults to false.
 
 =item wrap_nodes
 
-This parameter, if set to true, will wrap the node text.
+=item wrap_node_labels
 
-Due to an obscure bug that has been difficult to track down, this parameter
-sometimes causes graphs to not display properly. Use with caution.
+This parameter, if set to true, will wrap the node label text. This can be
+useful if the label is long. The line length is determined by the
+C<wrap_length> parameter. Defaults to false.
+
+=item wrap_edge_labels
 
 =item wrap_labels
 
-This parameter, if set to true, will wrap the label text. This property is
-always safe to use.
+This parameter, if set to true, will wrap the edge text. This can be useful if
+the rule message is long. The line length is determined by the C<wrap_length>
+parameter. Defaults to false C<wrap_labels> is deprecated and will be removed
+in a future version.
+
+=item text_wrap
+
+=item wrap_length
+
+The line length to use for wrapping text when C<wrap_nodes> or C<wrap_labels>
+is set to true. C<text_wrap> is deprecated and will be removed in a future
+version. Defaults to 25.
+
+=item node_params
+
+A hash reference of parameters to be passed to the GraphViz C<add_node()>
+method when seting up a state as a node. Only the C<label> parameter will be
+ignored. See the C<GraphViz|GraphViz/"add_node"> documentation for the list
+of supported paramters.
+
+=item edge_params
+
+A hash reference of parameters to be passed to the GraphViz C<add_node()>
+method when seting up a state as a node. See the
+C<GraphViz|GraphViz/"add_edge"> documentation for the list of supported
+paramters.
 
 =back
 
-B<Note:> If C<GraphViz> is not available on your system, this method will warn
-and return.
+B<Note:> If either C<GraphViz> or C<Text::Wrap> is not available on your
+system, C<graph()> will simply will warn and return.
 
 =cut
 
 sub graph {
     my $self = shift;
+    my $params = ref $_[0] ? shift : {};
+
     eval "use GraphViz 2.00; use Text::Wrap";
     if ($@) {
         warn "Cannot create graph object: $@";
         return;
     }
-    my $params = ref $_[0] ? shift : {};
-    $Text::Wrap::columns = $params->{text_wrap} || 25;
-    my $machine = clone($machines{$self}->{graph});
+
+    # Handle backwards compatibility.
+    $params->{wrap_node_labels} = $params->{wrap_nodes}
+        unless exists $params->{wrap_node_labels};
+    $params->{wrap_edge_labels} = $params->{wrap_labels}
+        unless exists $params->{wrap_edge_labels};
+    $params->{wrap_length} = $params->{text_wrap}
+        unless exists $params->{wrap_length};
+
+    # Set up defaults.
+    local $Text::Wrap::columns = $params->{wrap_length} || 25;
+    my @node_params = %{ $params->{node_params} || {} };
+    my @edge_params = %{ $params->{edge_params} || {} };
+
+    # Iterate over the states.
+    my $machine = $machines{$self};
     my $graph = GraphViz->new(@_);
-    while (my ($state, $definition) = splice @$machine => 0, 2) {
-        my $node = $params->{wrap_nodes}
-            ? wrap('','',$state)
-            : $state;
-        $graph->add_node($node);
-        next unless exists $definition->{rules};
-        while (my ($rule, $condition) = splice @{$definition->{rules}} => 0, 2) {
-            my @edge = ($state => $rule);
-            if (ref $condition eq 'HASH' && exists $condition->{message}) {
-                my $label = $params->{wrap_labels}
-                    ? wrap('','',$condition->{message})
+    for my $state (@{ $machine->{ord} }) {
+        my $def = $states{$state};
+        my $name  = $def->{name};
+
+        my $label = !$def->{label} ? $name
+            : $params->{with_state_name}  ? "$name\n\n$def->{label}"
+            :                               $def->{label};
+
+        $graph->add_node(
+            $name,
+            @node_params,
+            label => $params->{wrap_node_labels} ? wrap('', '', $label) : $label,
+        );
+        next unless exists $def->{rules};
+        for my $condition (@{ $def->{rules} }) {
+            my $rule = $condition->{state}->name;
+            my @edge = ($name => $rule);
+            if ($condition->{message}) {
+                push @edge, label => $params->{wrap_edge_labels}
+                    ? wrap('', '', $condition->{message})
                     : $condition->{message};
-                push @edge => 'label', $label;
             }
-            $graph->add_edge(@edge, decorate => 1);
+            $graph->add_edge( @edge, @edge_params );
         }
     }
     return $graph;
@@ -1053,6 +1108,12 @@ leaks.
 
 =cut
 
+# This method deletes the record from %machines, which has a reference to each
+# state, so those are deleted too. Each state refers back to the FSA::Rules
+# object itself, so as each of them is destroyed, it's removed from %states
+# and the FSA::Rules object gets all of its references defined in this file
+# freed, too. No circular references, so no problem.
+
 sub DESTROY { delete $machines{+shift}; }
 
 ##############################################################################
@@ -1078,7 +1139,7 @@ sub STORABLE_freeze {
     my ($self, $clone) = @_;
     return if $clone;
     my $fsa = $machines{$self};
-    return ( "$self", [ { %$self }, $fsa, @states{ @{ $fsa->{ord} } } ] );
+    return ( $self, [ { %$self }, $fsa, @states{ @{ $fsa->{ord} } } ] );
 }
 
 ##############################################################################
@@ -1090,13 +1151,12 @@ sub STORABLE_freeze {
 =cut
 
 sub STORABLE_thaw {
-    my ($self, $clone, $addr, $data) = @_;
+    my ($self, $clone, $junk, $data) = @_;
     return if $clone;
     %{ $self }                  = %{ shift @$data };
     my $fsa                     = shift @$data;
     $machines{ $self }          = $fsa;
-    $addr                       = "$self";
-    @states{ @{ $fsa->{ord} } } = map { $_->{machine} = $addr; $_ } @$data;
+    @states{ @{ $fsa->{ord} } } = @$data;
     return $self;
 }
 
@@ -1151,6 +1211,18 @@ sub name { $states{shift()}->{name} }
 
 ##############################################################################
 
+=head3 label
+
+  my $label = $state->label;
+
+Returns the label of the state.
+
+=cut
+
+sub label { $states{shift()}->{label} }
+
+##############################################################################
+
 =head3 machine
 
   my $machine = $state->machine;
@@ -1159,7 +1231,7 @@ Returns the FSA::Rules object for which the state was defined.
 
 =cut
 
-sub machine { $machines{ $states{shift()}->{machine} }->{self} }
+sub machine { $states{shift()}->{machine} }
 
 ##############################################################################
 
